import { Connection } from "@grapecity-software/js-collaboration-client";

/**
 * The interface of ot type, user can customize the ot behavior of the collaboration.
 * @template S The type of snapshot data.
 * @template T The type of op data.
 */
export interface OT_Type<S = unknown, T = unknown> {
    /**
     * The type uri of the ot type.
     */
    uri: string;
    /**
     * Create the snapshot data from op data.
     * @param data The data of the snapshot.
     * @returns The snapshot data.
     */
    create?(data: S): S;
    /**
     * Transform op1, to handling conflicts.
     * @param op1 The op to be transformed.
     * @param op2 The basis of the operation being transformed (op1 is transformed based on op2).
     * @param side Indicates which parameter's corresponding op arrives at the server later.
     * @returns The transformed op1.
     */
    transform(op1: T, op2: T, side: 'left' | 'right'): T;
    /**
     * Transform op1 and op2.
     * @param op1 The op to be transformed based on op2.
     * @param op2 The op to be transformed based on op1.
     * @returns The transformed op1 and op2.
     */
    transformX?(op1: T, op2: T): T[];
    /**
     * Apply op to snapshot.
     * @param snapshot
     * @param op
     * @returns The snapshot after applying op.
     */
    apply?(snapshot: S, op: T): S;
    /**
     * Normalize op.
     * @param op 
     * @returns The normalized op.
     */
    normalize?(op: T): T;
    /**
     * Compose two ops into one.
     * @param op1 The op to be composed with op2.
     * @param op2 The op to be composed with op1.
     * @returns The composed op.
     */
    compose?(op1: T, op2: T): T;
    /**
     * Create client doc data from snapshot data.
     * @param snapshot
     * @returns The deserialized snapshot data.
     */
    deserialize?(snapshot: S): S;
    /**
     * Create client doc data from op data.
     * @param data
     * @returns The deserialized snapshot data.
     */
    createDeserialized?(data: S): S | undefined;
}

/**
 * The types manager for collaboration, provide a way to register ot type and get types through type uri.
 */
export class TypesManager {
    /**
     * Register ot type.
     * @param {OT_Type} type The type object which is user-defined.
     */
    static register<S, T> (type: OT_Type<S, T>): void;
    /**
     * Get ot type through the type uri.
     * @param {string} typeUri The type uri.
     * @return {OT_Type} The type object of specified type uri.
     */
    static getType<S, T> (typeUri: string): OT_Type<S, T>;
}

/**
 * The shared doc object to request snapshot and ops.
 * @template S The type of snapshot data.
 * @template T The type of op data.
 */
export declare class SharedDoc<S = unknown, T = unknown> {
    /**
     * The version of the snapshot.
     */
    version?: number;
    /**
     * The ot type of the snapshot.
     */
    type?: OT_Type<S, T>;
    /**
     * The snapshot data.
     */
    data?: S;
    constructor(conn: Connection);
    /**
     * The id of the shared doc.
     * @readonly
     * @returns {string} The id of the shared doc.
     */
    get id(): string;
    /**
     * The connection of the shared doc.
     * @readonly
     * @returns {Connection} The connection of the shared doc.
     */
    get connection(): Connection;
    /**
     * Fetch the snapshot data.
     * @param callback The callback function which will be called after the snapshot is fetched, the fetched snapshot data is `doc.data`.
     */
    fetch(callback?: ICallBack<void>): void;
    /**
     * Subscribe the snapshot data.
     * @param callback The callback function which will be called after the snapshot is subscribed, the subscribed snapshot data is `doc.data`.
     */
    subscribe(callback?: ICallBack<void>): void;
    /**
     * Submit an op to the document.
     * @param component The op data.
     * @param options The options of the op.
     * @param callback The callback function which will be called after the op is submitted. 
     */
    submitOp(component: T, options?: IOptions, callback?: ICallBack<void>): void;
    /**
     * Create the document.
     * @param data The op data.
     * @param type The ot type.
     * @param options The options of the op.
     * @param callback The callback function which will be called after the document is created.
     */
    create(data: S, type: string, options: IOptions, callback?: ICallBack<void>): void;
    /**
     * Fetch the history snapshot data.
     * @param options The options of the history snapshot.
     * @param callback The callback function which will be called after the history snapshot is fetched.
     */
    fetchHistorySnapshot(options: IHistorySnapshotRequestOptions, callback: ICallBack<ISnapshot<S>>): void;
    /**
     * Delete the document.
     * @param options The options of the op.
     * @param callback The callback function which will be called after the document is deleted.
     */
    del(options: IOptions, callback: ICallBack<void>): void;
    /**
     * Hard rollback the snapshot data to the latest version.
     */
    hardRollback(): void;
    /**
     * Register a listener for an event.
     * @param name - The event name.
     * @param f - The event handler.
     */
    on<NAME extends keyof IEvents<S, T> & string>(name: NAME, f: IEvents<S, T>[NAME]): IEvents<S, T>[NAME];
    /**
     * Register a listener for an event that will only be called once.
     * @param name - The event name.
     * @param f - The event handler.
     */
    once<NAME_1 extends keyof IEvents<S, T> & string>(name: NAME_1, f: IEvents<S, T>[NAME_1]): void;
    /**
     * Remove a listener for an event.
     * @param name - The event name.
     * @param f - The event handler.
     */
    off<NAME_2 extends keyof IEvents<S, T> & string>(name: NAME_2, f: IEvents<S, T>[NAME_2]): void;
    /**
     * Destroy the observable.
     */
    destroy(): void;
}

/**
 * The events of the SharedDoc.
 */
export interface IEvents<S, T> {
    /**
     * Error occurred.
     */
    error: (err: OTError) => void;
    /**
     * After client fetch or subscribe the snapshot.
     */
    load: () => void;
    create: (source?: unknown) => void;
    del: (data: S, source?: unknown) => void;
    hardRollback: () => void;
    /**
     * Before apply op.
     */
    beforeOp: (op: T, source?: unknown, src?: string) => void;
    /**
     * After apply op.
     */
    op: (op: T, source?: unknown, src?: string) => void;
}

export interface IHistorySnapshotRequestOptions {
    version?: number;
}

export interface ICallBack<T = unknown> {
    (err?: OTError | null, data?: T): void;
}

export interface IOp<T = unknown> {
    v: number;
    create?: ICreateComponent;
    op?: T;
    del?: boolean;
}
export interface ICreateComponent {
    type: string;
    data: unknown;
}

export interface IRequestMessage<T> {
    /**
     * message action
     */
    a: MessageActions;
    /**
     * version
     */
    v?: number;
    create?: ICreateComponent;
    op?: T;
    del?: boolean;
    error?: IErrorData;
}
export interface IReplyMessage<S, T> {
    a: string;
    v?: number;
    create?: ICreateComponent;
    op?: T;
    del?: boolean;
    error?: IErrorData;
    data?: ISnapshot<S>;
}

export interface IIdToError {
    [docId: string]: string | OTError;
}

/**
 * Message actions.
 */
export enum MessageActions {
    fetch,
    fetchOps,
    fetchHistorySnapshot,
    subscribe,
    unsubscribe,
    op,
}

export class OTError extends Error {
    static CODES: Record<string, string>;

    public name: string;
    public data?: unknown;
    public idToError?: IIdToError | null;
}
/**
 * Error codes of ot errors.
 */
export enum ERROR_CODES {
    ERR_DOC_ALREADY_CREATED,
    ERR_DOC_DOES_NOT_EXIST,
    ERR_DOC_TYPE_NOT_RECOGNIZED,
    ERR_DOC_WAS_DELETED,

    ERR_MAX_SUBMIT_RETRIES_EXCEEDED,
    ERR_MAX_SUBMIT_SNAPSHOT_RETRIES_EXCEEDED,
    ERR_MESSAGE_BADLY_FORMED,
    ERR_OP_ALREADY_SUBMITTED,

    ERR_HARD_ROLLBACK_FETCH_FAILED,

    ERR_OT_OP_NOT_APPLIED,
    ERR_OT_OP_TRANSFORM_FAILED,

    ERR_SNAPSHOT_READS_REJECTED,
    ERR_SUBMIT_TRANSFORM_OPS_NOT_FOUND,
    ERR_UNKNOWN_ERROR
}
export interface IErrorData {
    code: number;
    message: string;
}

export interface IOptions {
    source?: string;
}

export interface ISnapshot<S = unknown> {
    id: string;
    v: number;
    type?: string;
    data?: S;
}